#include <eyebot.h>

#define SLAVES 5
#define SSIZE  4096

/* pointer to thread control blocks */
struct tcb *slaveBlocks[SLAVES], *masterBlock;

/* semaphores */
struct sem semaphores[SLAVES];

/* Code for the slave processes.  Each process acquires its UID at first, and uses
 * it here to differentiate activity.
 */
void slave() {
	int id;

	/* acquire UID (supplied by main during OSSpawn operation) */
	id = OSGetUID(0);
	LCDPrintf("Slave %d starting ...\n", id);

	//switch (id) {

	while(1) {
		OSSemP(&semaphores[id]);                /* occupy semaphore */

		//task code goes here

		OSSemV(&semaphores[id]);                /* free semaphore */
		OSSleep(70);                            /* sleep 0.7s --> other threads */
		//OSReschedule instead here?
	}

	//} //end switch

}



/* Code for master process.  Master handles user input via keys to terminate tasks
 * for testing purposes.
 */
void master() {
	int i;

  /** start with both semaphores and slave tasks blocked */
  int block[SLAVES] = {1,1,1};
  LCDPrintf("master start\n");

  /** display appropr. menu for soft-keys, depend. on slave task status */
  LCDMenu("V.0", "V.1", "V.2", "END");

  while(1) {
    switch (KEYRead())
    {
      case KEY1: if (block[0]) { OSSemV(&sema[0]); LCDMenu("P.0", "","",""); }
			  else { OSSemP(&sema[0]); LCDMenu("V.0", "","",""); }
		 block[0] = !block[0];
		 break;

      case KEY2: if (block[1]) { OSSemV(&sema[1]); LCDMenu("", "P.1", "",""); }
			  else { OSSemP(&sema[1]); LCDMenu("", "V.1", "",""); }
		 block[1] = !block[1];
		 break;

      case KEY3: if (block[2]) { OSSemV(&sema[2]); LCDMenu("","", "P.2", ""); }
			  else { OSSemP(&sema[2]); LCDMenu("","", "V.2", ""); }
		 block[2] = !block[2];
		 break;

      case KEY4: /** kill both slaves and then exit master itself */
		 for (i=0; i<SLAVES; i++) OSKill(slave_p[i]);
		 OSKill(0);
		 break;
    }
    OSReschedule();  /* transfer control to other threads */
  } /* end while */
} /* end master */



/* Main of demo program */
int main() {
	int i;

	/* initialise multithreading */
	LCDPrintf("MT init ... ");
	OSMTInit(COOP);
	LCDPrintf("done!\n");

	/* initialise semaphores */
	LCDPrintf("Sem init ... ");
	for (i=0; i<SLAVES; i++) OSSemInit(&semaphores[i],0);
	LCDPrintf("done!\n");

	/* initialise threads:
	 * <initialised control block>
	   OSSpawn(<thread name>,<control block pointer>,<size>,<priority>,<uid>)
	 */
	LCDPrintf("Spawn init ...\n");
	LCDPrintf("Left PSD ... ");
	leftPSDCode = OSSpawn("LeftPSD", leftPSD, SSIZE, MIN_PRI, 0);
	if(!leftPSDCode) OSPanic("failed!\n");
	LCDPrintf("done!\n");

	LCDPrintf("Right PSD ... ");
	slaveBlocks[1] = OSSpawn("RightPSD", rightPSD, SSIZE, MIN_PRI, 1);
	if(!slaveBlocks[1]) OSPanic("failed!\n");
	LCDPrintf("done!\n");

	LCDPrintf("Camera ... ");
	slaveBlocks[2] = OSSpawn("Camera", camera, SSIZE, MIN_PRI, 2);
	if(!slaveBlocks[2]) OSPanic("failed!\n");
	LCDPrintf("done!\n");

	LCDPrintf("LCD Output ... ");
	slaveBlocks[3] = OSSpawn("LCDOut", lcdOut, SSIZE, MIN_PRI, 3);
	if(!slaveBlocks[0]) OSPanic("failed!\n");
	LCDPrintf("done!\n");


/** READY all threads and start one */
for (i=0; i<SLAVES; i++) OSReady(slave_p[i]);
OSReady(master_p);
OSReschedule();

/* ----------------------------------------------------------------------- */
/** processing will return HERE, when there is no READY thread left */

LCDPrintf("back to main\n");
LCDMenu (" ", " ", " ", "END");
KEYWait(KEY4);
return 0;
}

