#include <eyebot.h>
#include <string.h>

#define TASKS 		6	//6 in full version
#define TASK_SIZE  	16384
#define CAMERA_SIZE	32768
#define QUEUE_SIZE 	16
#define RESET_VAL	128

/* type definition for output queue items */
typedef struct {
	int x;
	int y;
	char text[16];
} QueueItem;

/* pointer to thread control blocks */
struct tcb *codeBlocks[TASKS];

/* semaphores */
struct sem queueSem;
struct sem lcdSem;
int currQueueTask;
int currLCDTask;

/* pointers to current queue front and back */
int queueHead = 0;
int queueTail = 0;
QueueItem queue[QUEUE_SIZE];

/* integers representing status of basic tasks */
int leftPSDStatus = RESET_VAL;
int rightPSDStatus = RESET_VAL;
int lcdOutStatus = RESET_VAL;
int cameraStatus = RESET_VAL;
int taskStates[4] = {1,1,1,1};

/* assorted globals */
PSDHandle lPSD;
PSDHandle rPSD;
image frame;

/* augmented semaphore acquisition */
void GetQueue() {
	OSSemP(&queueSem);
	currQueueTask = OSGetUID(0);
}

void GetLCD() {
	OSSemP(&lcdSem);
	currLCDTask = OSGetUID(0);
}

/* safe lcd printout */
void SafeLCDPrintf(char* text, int row) {
	GetLCD();
	LCDSetPos(row,0);
	LCDPrintf(text);
	OSSemV(&lcdSem);
}

/* exclusive writing to the output queue */
int queueWrite(int x, int y, char *t) {

	GetQueue();
	
	if ((queueTail++)%QUEUE_SIZE == queueHead) { //queue is Full
		OSSemV(&queueSem);
		return 0;
	}
	else {
		//write struct accordingly
		queue[queueTail].x = x;
		queue[queueTail].y = y;
		strcpy((queue[queueTail]).text,t);
		queueTail = (queueTail++)%QUEUE_SIZE;
		OSSemV(&queueSem);
		return 1;
	}
}

/* Code for the concurrent tasks */
void leftPSD() {
	int psdValue = 0;
	char text[16];

	while(1) {
		leftPSDStatus = RESET_VAL;
		//get PSD value
		psdValue = PSDGet(lPSD);

		sprintf(text,"%d",psdValue);

		//write output queue
		while (!queueWrite(0,11,text));

		OSSleep(1);
	}
}

void rightPSD() {
	int psdValue = 0;
	char text[16];

	while(1) {
		rightPSDStatus = RESET_VAL;
		//get PSD value
		psdValue = PSDGet(rPSD);

		sprintf(text, "%d",psdValue);
		
		//write output queue
		while (!queueWrite(1,11,text));

		OSSleep(1);
	}
}

void camera() {
	int i;
	float average = 0;
	int value = 0;
	char text[16];
	
	while(1) {
		cameraStatus = RESET_VAL;
		//Obtain greyscale image
		CAMGetFrame((image *)&frame);
	
		//Calculate average of main diagonal
		for (i = 0; i < 60; i++) {
			average *= i;
			average += (float)((BYTE)(frame[i][i]));
			average /= (i+1);
		}
		
		value = (int)average;
		sprintf(text, "%d",value);

		while (!queueWrite(2,11,text));
		
		GetLCD();
		LCDPutGraphic((image *)&frame);
		OSSemV(&lcdSem);		//free semaphore
		
		OSSleep(1);
	}
}

void lcdOut() {
	int id;
	int keyPressed;//for test version only
	int x, y;
	char text[16];
	
	while(1) {
		lcdOutStatus = RESET_VAL;

		GetQueue();	//occupy queue semaphore

		if (queueHead == queueTail) {//queue is Empty
			OSSemV(&queueSem);
		}
		else {
			x = queue[queueHead].x;	//copy queue item
			y = queue[queueHead].y;
			strcpy(text, queue[queueHead].text);

			queueHead = (queueHead+1)%QUEUE_SIZE;	//update queue
			OSSemV(&queueSem);

			GetLCD();
			LCDSetPos(x,y);	//output data
			LCDPutString(text);
			OSSemV(&lcdSem);
		}
		OSSleep(1);
	}
}

//Restarts tasks when they are killed
void watchdog() {
	char statusText[16];
	
	while(1) {
		leftPSDStatus--;
		rightPSDStatus--;
		lcdOutStatus--;
		cameraStatus--;
		
		if (leftPSDStatus <= 0) {
			taskStates[0] = 0;
			/*OSForbid();
			//OSKill(codeBlocks[0]);
			codeBlocks[0] = OSSpawn("LeftPSD", leftPSD, TASK_SIZE, MIN_PRI, 0);
			leftPSDStatus = RESET_VAL;
			if (currQueueTask == 0) OSSemInit(&queueSem,1);
			if (currLCDTask == 0) OSSemInit(&lcdSem,1);
			OSReady(codeBlocks[0]);
			OSPermit();
			OSWait(200);*/
		}
		else taskStates[0] = 1;
		if (rightPSDStatus <= 0) {
			taskStates[1] = 0;
			/*OSForbid();
			//OSKill(codeBlocks[1]);
			codeBlocks[1] = OSSpawn("RightPSD", rightPSD, TASK_SIZE, MIN_PRI, 1);
			rightPSDStatus = RESET_VAL;
			if (currQueueTask == 1) OSSemInit(&queueSem,1);
			if (currLCDTask == 1) OSSemInit(&lcdSem,1);
			OSReady(codeBlocks[1]);
			OSPermit();
			OSWait(200);*/
		}
		else taskStates[1] = 1;
		if (cameraStatus <= 0) {
			taskStates[2] = 0;
			/*OSForbid();
			//OSKill(codeBlocks[2]);
			codeBlocks[2] = OSSpawn("Camera", camera, CAMERA_SIZE, MIN_PRI+1, 2);
			lcdOutStatus = RESET_VAL;
			if (currQueueTask == 2) OSSemInit(&queueSem,1);
			if (currLCDTask == 2) OSSemInit(&lcdSem,1);
			OSReady(codeBlocks[2]);
			OSPermit();
			OSait(200);*/
		}
		else taskStates[2] = 1;
		if (lcdOutStatus <= 0) {
			taskStates[3] = 0;
			/*OSForbid();
			//OSKill(codeBlocks[3]);
			codeBlocks[3] = OSSpawn("LCDOut", lcdOut, TASK_SIZE, MIN_PRI+1, 3);
			cameraStatus = RESET_VAL;
			if (currQueueTask == 3) OSSemInit(&queueSem,1);
			if (currLCDTask == 3) OSSemInit(&lcdSem,1);
			OSReady(codeBlocks[3]);
			OSPermit();
			OSWait(200);*/
		}
		else taskStates[3] = 1;
		
		sprintf(statusText, "%d %d %d %d", taskStates[0], taskStates[1], taskStates[2], taskStates[3]);
		SafeLCDPrintf(statusText,7);
		
		OSSleep(1);
	}

}

void terminator() {
	int keyPressed;

	while(1) {
		keyPressed = KEYRead();

		switch (keyPressed) {
			case KEY1: {
				if (taskStates[0]) OSKill(codeBlocks[0]);
				break;
			};
			case KEY2: {
				if (taskStates[1]) OSKill(codeBlocks[1]);
				break;
			};
			case KEY3: {
				if (taskStates[2]) OSKill(codeBlocks[2]);
				break;
			};
			case KEY4: {
				if (taskStates[3]) OSKill(codeBlocks[3]);
				break;
			};
			default: break;
		}
		//Read key input and terminate tasks 0-3 if indicated
		OSSleep(1);
	}
}


/* Main program */
int main() {
	int i, keyPressed;

	//Initialise PSDs
	PSDRelease();
	
	lPSD = PSDInit(PSD_LEFT);
	rPSD = PSDInit(PSD_RIGHT);

	PSDStart(lPSD,1);
	PSDStart(rPSD,1);

	//Initialise camera
	CAMInit(NORMAL);
	
	/* initialise multithreading */
	LCDPrintf("MUL ... ");
	OSMTInit(PREEMPT);
	LCDPrintf("done!\n");

	/* initialise semaphores */
	LCDPrintf("SEM ... ");
	OSSemInit(&queueSem,1);
	OSSemInit(&lcdSem,1);
	LCDPrintf("done!\n");

	/* initialise threads:
	 * <initialised control block>
	   OSSpawn(<thread name>,<control block pointer>,<size>,<priority>,<uid>)
	 */
	LCDPrintf("SPN ...\n");
	LCDPrintf("LPS ... ");
	codeBlocks[0] = OSSpawn("LeftPSD", leftPSD, TASK_SIZE, MIN_PRI, 0);
	OSWait(50);
	
	if(!codeBlocks[0]) OSPanic("failed!\n");
	LCDPrintf("done!\n");

	OSWait(50);
	
	LCDPrintf("RPS ... ");
	codeBlocks[1] = OSSpawn("RightPSD", rightPSD, TASK_SIZE, MIN_PRI, 1);
	OSWait(50);
	
	if(!codeBlocks[1]) OSPanic("failed!\n");
	LCDPrintf("done!\n");

	OSWait(50);
	
	LCDPrintf("CAM ... ");
	codeBlocks[2] = OSSpawn("Camera", camera, CAMERA_SIZE, MAX_PRI, 2);
	OSWait(50);
	
	if(!codeBlocks[2]) OSPanic("failed!\n");
	LCDPrintf("done!\n");

	OSWait(50);
	
	LCDPrintf("LCD ... ");
	codeBlocks[3] = OSSpawn("LCDOut", lcdOut, TASK_SIZE, MIN_PRI+1, 3);
	OSWait(50);
	
	if(!codeBlocks[3]) OSPanic("failed!\n");
	LCDPrintf("done!\n");
	
	OSWait(50);
	
	LCDPrintf("WAD ... ");
	codeBlocks[4] = OSSpawn("Watchdog", watchdog, TASK_SIZE, MIN_PRI, 4);
	OSWait(50);
	
	if(!codeBlocks[4]) OSPanic("failed!\n");
	LCDPrintf("done!\n");

	OSWait(50);
	
	LCDPrintf("TRM ... ");
	codeBlocks[5] = OSSpawn("Terminator", terminator, TASK_SIZE, MIN_PRI, 5);
	OSWait(50);

	if(!codeBlocks[5]) OSPanic("failed!\n");
	LCDPrintf("done!\n");

	LCDClear();
	
	//Make all threads ready, start one
	
	for (i = 0; i < TASKS; i++) OSReady(codeBlocks[i]);
	OSPermit();
	OSWait(200); //startup time
	
	//Return here when there are no ready threads left */
	LCDPrintf("Ready to close!\n");
	LCDMenu (" ", " ", " ", "END");
	KEYWait(KEY4);

	PSDRelease();
	CAMRelease();
	
	return 0;
}

