#include <eyebot.h>
#include <string.h>

#define TASKS 		3	//6 in full version
#define TASK_SIZE  	4096
#define QUEUE_SIZE 	512

/* type definition for output queue items */
typedef struct {
	int x;
	int y;
	char text[16];
} QueueItem;

/* pointer to thread control blocks */
struct tcb *codeBlocks[TASKS];

/* semaphores */
struct sem queueSem;
struct sem lcdSem;

/* pointers to current queue front and back */
int queueHead;
int queueTail;
QueueItem queue[QUEUE_SIZE];

/* assorted globals */
PSDHandle lPSD;
PSDHandle rPSD;

/* exclusive writing to the output queue */
int queueWrite(int x, int y, char *t) {
	OSSemP(&queueSem);

	if ((queueTail++)%QUEUE_SIZE == queueHead) { //queue is Full
		OSSemV(&queueSem);
		return 0;
	}
	else {
		//write struct accordingly
		queue[queueTail].x = x;
		queue[queueTail].y = y;
		strcpy((queue[queueTail]).text,t);
		queueTail = (queueTail++)%QUEUE_SIZE;
		OSSemV(&queueSem);
		return 1;
	}
}


/* Code for the concurrent tasks */
void leftPSD() {
	int id;
	int psdValue = 0;
	char text[16];

	/* acquire UID (supplied by main during OSSpawn operation) */
	id = OSGetUID(0);

	while(1) {
		//get PSD value
		psdValue = PSDGet(lPSD);

		sprintf(text,"LPSD: %d",psdValue);

		//write output queue
		while (!queueWrite(0,0,text)) OSReschedule();

		//OSReschedule();
		OSSleep(70);
	}
}

void rightPSD() {
	int id;
	int psdValue = 0;
	char text[16];

	/* acquire UID (supplied by main during OSSpawn operation) */
	id = OSGetUID(0);

	while(1) {
		//get PSD value
		psdValue = PSDGet(rPSD);

		sprintf(text, "RPSD: %d",psdValue);

		//write output queue
		while (!queueWrite(0,1,text)) OSReschedule();

		//OSReschedule();
		OSSleep(70);
	}
}

/*void camera() {
	int id;

	//acquire UID (supplied by main during OSSpawn operation)
	id = OSGetUID(0);
	//LCDPrintf("Camera, ID=%d...\n",id);

	while(1) {
		OSSemP(&queueSem);		//occupy semaphore
		OSSemP(&lcdSem);
		//acquire camera image, process, add result to output queue
		//display camera image on LCD
		OSSemV(&queueSem);
		OSSemV(&lcdSem);		//free semaphore
		OSReschedule();
	}
}*/

void lcdOut() {
	int id;
	int keyPressed;//for test version only
	int x, y;
	char text[16];

	/* acquire UID (supplied by main during OSSpawn operation) */
	id = OSGetUID(0);
	//LCDPrintf("LCD Output, ID=%d...\n",id);

	while(1) {
		//test version only!!
		keyPressed = KEYRead();
		if (keyPressed == KEY4) OSPanic("Halted!");

		OSSemP(&queueSem);	//occupy queue semaphore

		if (queueHead == queueTail) //queue is Empty
			OSSemV(&queueSem);
		else {
			x = queue[queueHead].x;	//copy queue item
			y = queue[queueHead].y;
			strcpy(text, queue[queueHead].text);

			queueHead = (queueHead++)%QUEUE_SIZE;	//update queue
			OSSemV(&queueSem);

			OSSemP(&lcdSem);
			LCDSetPos(x,y);	//output data
			LCDPutString(text);
			OSSemV(&lcdSem);
		}
		//OSReschedule();
		OSSleep(70);
	}
}

/*void watchdog() {
	int id;

	//acquire UID (supplied by main during OSSpawn operation)
	id = OSGetUID(0);
	LCDPrintf("Watchdog, ID=%d...\n",id);

	while(1) {
		OSSemP(&semaphores[id]);		//occupy semaphore

		//check status of PSD, camera, and output tasks, and print
		//status if they have stopped running

		OSSemV(&semaphores[id]);		//free semaphore
		OSReschedule();
	}

}*/

/*void terminator() {
	int id, keyPressed;

	//acquire UID (supplied by main during OSSpawn operation)
	id = OSGetUID(0);
	LCDPrintf("Terminator, ID=%d...\n",id);

	while(1) {
		OSSemP(&semaphores[id]);		//occupy semaphore

		//use OSKill(struct tcb *thread)
		keyPressed = KEYRead();

		switch (keyPressed) {
			case KEY1: {
				OSKill(codeBlocks[0]);
				break;
			};
			case KEY2: {
				OSKill(codeBlocks[1]);
				break;
			};
			case KEY3: {
				OSKill(codeBlocks[2]);
				break;
			};
			case KEY4: {
				OSKill(codeBlocks[3]);
				break;
			};
			default: break;
		}
		//Read key input and terminate tasks 0-3 if indicated

		OSSemV(&semaphores[id]);		//free semaphore
		OSReschedule();
	}
}*/


/* Code for master process.  Master handles user input via keys to terminate tasks
 * for testing purposes.

void master() {
	int i;

  //start with both semaphores and slave tasks blocked
  int block[SLAVES] = {1,1,1};
  LCDPrintf("master start\n");

  //display appropr. menu for soft-keys, depend. on slave task status
  LCDMenu("V.0", "V.1", "V.2", "END");

  while(1) {
    switch (KEYRead())
    {
      case KEY1: if (block[0]) { OSSemV(&sema[0]); LCDMenu("P.0", "","",""); }
			  else { OSSemP(&sema[0]); LCDMenu("V.0", "","",""); }
		 block[0] = !block[0];
		 break;

      case KEY2: if (block[1]) { OSSemV(&sema[1]); LCDMenu("", "P.1", "",""); }
			  else { OSSemP(&sema[1]); LCDMenu("", "V.1", "",""); }
		 block[1] = !block[1];
		 break;

      case KEY3: if (block[2]) { OSSemV(&sema[2]); LCDMenu("","", "P.2", ""); }
			  else { OSSemP(&sema[2]); LCDMenu("","", "V.2", ""); }
		 block[2] = !block[2];
		 break;

      case KEY4: //kill both slaves and then exit master itself
		 for (i=0; i<SLAVES; i++) OSKill(slave_p[i]);
		 OSKill(0);
		 break;
    }
    OSReschedule();  //transfer control to other threads
  }
} */



/* Main of demo program */
int main() {
	int i;

	lPSD = PSDInit(PSD_LEFT);
	rPSD = PSDInit(PSD_RIGHT);

	/* initialise multithreading */
	LCDPrintf("MT init ... ");
	OSMTInit(COOP);
	LCDPrintf("done!\n");

	/* initialise semaphores */
	LCDPrintf("Sem init ... ");
	OSSemInit(&queueSem,0);
	OSSemInit(&lcdSem,0);
	LCDPrintf("done!\n");

	/* initialise threads:
	 * <initialised control block>
	   OSSpawn(<thread name>,<control block pointer>,<size>,<priority>,<uid>)
	 */
	LCDPrintf("Spawn init ...\n");
	LCDPrintf("Left PSD ... ");
	codeBlocks[0] = OSSpawn("LeftPSD", leftPSD, TASK_SIZE, MIN_PRI, 0);
	if(!codeBlocks[0]) OSPanic("failed!\n");
	LCDPrintf("done!\n");

	LCDPrintf("Right PSD ... ");
	codeBlocks[1] = OSSpawn("RightPSD", rightPSD, TASK_SIZE, MIN_PRI, 1);
	if(!codeBlocks[1]) OSPanic("failed!\n");
	LCDPrintf("done!\n");

/*	LCDPrintf("Camera ... ");
	codeBlocks[2] = OSSpawn("Camera", camera, TASK_SIZE, MIN_PRI, 2);
	if(!codeBlocks[2]) OSPanic("failed!\n");
	LCDPrintf("done!\n");*/

	LCDPrintf("LCD Output ... ");
	codeBlocks[2] = OSSpawn("LCDOut", lcdOut, TASK_SIZE, MIN_PRI, 3);
	if(!codeBlocks[2]) OSPanic("failed!\n");
	LCDPrintf("done!\n");

/*	LCDPrintf("Watchdog ...");
	codeBlocks[4] = OSSpawn("Watchdog", watchdog, TASK_SIZE, MIN_PRI, 4);
	if(!codeBlocks[4]) OSPanic("failed!\n");
	LCDPrintf("done!\n");

	LCDPrintf("Terminator ...");
	codeBlocks[5] = OSSpawn("Terminator", terminator, TASK_SIZE, MIN_PRI, 5);
	LCDPrintf("done!\n");*/

	//Make all threads ready, start one
	for (i = 0; i < TASKS; i++) OSReady(codeBlocks[i]);
	OSReschedule();

	//Return here when there are no ready threads left */
	LCDPrintf("Ready to close!\n");
	LCDMenu (" ", " ", " ", "END");
	KEYWait(KEY4);

	return 0;
}

