#include <eyebot.h>
#include "fuzzy.c"

//Frequency with which control routine is called
#define TIMESCALE 10
#define SPEED_INC 10

int desired = 0;//(ticks/time interval)

MotorHandle left;
MotorHandle right;

QuadHandle leftQuad;
QuadHandle rightQuad;

float leftError=0,oldLeftError=0,leftErrorDot=0;
float rightError=0,oldRightError=0,rightErrorDot=0;

int oldLeftVal=0;
int oldRightVal=0;

int time=0,oldTime=0,timeChange=0;

int leftMotorVal=0, rightMotorVal=0;

void controlFunc() {
	int leftVal;
	int rightVal;
	int leftChange;
	int rightChange;
	float leftDist;
	float rightDist;
	float leftVel;
	float rightVel;
	
	leftVal = QUADRead(leftQuad);
	rightVal = QUADRead(rightQuad);

	leftDist = leftVal - oldLeftVal;
	rightDist = rightVal - oldRightVal;
	
	oldTime = time;
	time = OSGetCount();
	timeChange = time - oldTime;
	
	leftVel = leftDist/((float)timeChange);
	rightVel = rightDist/((float)timeChange);
	
	oldLeftVal = leftVal;
	oldRightVal = rightVal;
	
	leftError = leftVel - desired;
	rightError = rightVel - desired;

	LCDPrintf("DA: %d %2.1f\n",desired,leftVel);
	
//	LCDPrintf("EED: %2.1f ",leftError);
	
	leftErrorDot = (leftError - oldLeftError)/((float)timeChange);
	rightErrorDot = (rightError - oldRightError)/((float)timeChange);

	LCDPrintf("%2.1f\n",leftErrorDot);
	
	oldLeftError = leftError;
	oldRightError = rightError;
	
	leftChange = getFuzzyOutput(leftError, leftErrorDot);
	rightChange = getFuzzyOutput(rightError, rightErrorDot);
	
	//LCDPrintf("LC: %d\n",leftChange);
	
	leftMotorVal += leftChange;
	rightMotorVal += rightChange;	
	
	leftMotorVal = (leftMotorVal < -100)?(-100):leftMotorVal;
	leftMotorVal = (leftMotorVal > 100)?(100):leftMotorVal;
	
	rightMotorVal = (rightMotorVal < -100)?(-100):rightMotorVal;
	rightMotorVal = (rightMotorVal > 100)?(100):rightMotorVal;
	
//	LCDPrintf("Left: %d\n",leftMotorVal);
	
	MOTORDrive(left, leftMotorVal);
	MOTORDrive(right, rightMotorVal);
}

int main(int argc, char **argv) {

	int keyPressed = KEY3;
	
	desired = 0;

	LCDPrintf("Release ... ");
	
	MOTORRelease(left);
	MOTORRelease(right);
	QUADRelease(leftQuad);
	QUADRelease(rightQuad);
	
	LCDPrintf("done.\n");

	LCDPrintf("Motors ... ");
	
	left = MOTORInit(LEFTMOTOR);
	right = MOTORInit(RIGHTMOTOR);
	
	MOTORDrive(left,50);
	MOTORDrive(right,50);
	
	LCDPrintf("done.\n");
	
	LCDPrintf("Quads ... ");
	
	leftQuad = QUADInit(QUAD_LEFT);
	rightQuad = QUADInit(QUAD_RIGHT);

	LCDPrintf("done.\n");
	
	OSWait(100);
	
	MOTORDrive(left,0);
	MOTORDrive(right,0);

	OSWait(100);
	
	LCDPrintf("Starting main!\n");
	
	OSWait(100);

	time = OSGetCount();
	
	while (keyPressed != KEY4) {
		keyPressed = KEYRead();

		if (keyPressed == KEY1) desired += SPEED_INC;
		if (keyPressed == KEY2) desired -= SPEED_INC;

		LCDClear();

		controlFunc();
//		OSWait(50);
	}

	
	LCDPrintf("Release ... ");
	
	MOTORRelease(left);
	MOTORRelease(right);
	QUADRelease(leftQuad);
	QUADRelease(rightQuad);
	
	LCDPrintf("done.\n");
	
	return 1;
	
}

